import 'dart:async';

import 'package:flutter/material.dart';

import 'swipe_indicator_widget.dart';

typedef void _OnTap(int index);

/// The swipe widget
///
/// ```
/// Widget build(BuildContext context) {
///   var images = <String>[
///     'assets/images/demo/swipe_1.png',
///     'assets/images/demo/swipe_2.png',
///     'assets/images/demo/swipe_3.png',
///   ];
///
///   return SwipeWidget(
///     height: calculateHeightByScale(context, 16 / 9),
///     indicatorTheme: SwipeIndicatorTheme(
///       activeColor: Colors.blue,
///       inactiveColor: Colors.white,
///       backgroundColor: Colors.blue.withAlpha(100),
///     ),
///     itemCount: images.length,
///     itemBuilder: (BuildContext context, int index) {
//        return Image.asset(images[index], fit: BoxFit.cover);
//      },
//      onTap: (int index) {
//        Log.v('Swipe $index');
//      },
///   );
/// }
/// ```
class SwipeWidget extends StatefulWidget {
  final double width;
  final double height;
  final bool autoplay;
  final int autoplayDelay;
  final int initialIndex;
  final int itemCount;
  final IndexedWidgetBuilder itemBuilder;
  final SwipeIndicatorTheme indicatorTheme;
  final _OnTap? onTap;
  final ValueChanged<int>? onIndexChanged;

  SwipeWidget({
    this.width = double.infinity,
    this.height = 200,
    this.autoplay = true,
    this.autoplayDelay = 5,
    this.initialIndex = 0,
    required this.indicatorTheme,
    required this.itemCount,
    required this.itemBuilder,
    this.onTap,
    this.onIndexChanged,
    Key? key,
  }) : super(key: key);

  @override
  State<StatefulWidget> createState() => _SwipeWidget();
}

class _SwipeWidget extends State<SwipeWidget> {
  PageController? _pageController;
  Timer? _timer;

  int _currentIndex = 0;

  @override
  void initState() {
    super.initState();
    _currentIndex = widget.initialIndex;
    _pageController = new PageController(initialPage: _currentIndex);

    if (widget.autoplay) {
      WidgetsBinding.instance?.addPostFrameCallback((timeStamp) {
        _timer = new Timer.periodic(
            Duration(milliseconds: widget.autoplayDelay * 1000), (value) {
          if (_currentIndex + 1 >= widget.itemCount) {
            _currentIndex = 0;
          } else {
            _currentIndex++;
          }

          _pageController!.animateToPage(_currentIndex,
              duration: Duration(milliseconds: 200), curve: Curves.ease);
          setState(() {});
        });
      });
    }
  }

  @override
  void dispose() {
    _timer?.cancel();
    _timer = null;

    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return Container(
      width: widget.width,
      height: widget.height,
      child: Stack(
        alignment: Alignment.center,
        children: [
          GestureDetector(
              onTap: () {
                if (widget.onTap != null) {
                  widget.onTap!(_currentIndex);
                }
              },
              child: PageView.builder(
                itemCount: widget.itemCount,
                itemBuilder: widget.itemBuilder,
                controller: _pageController,
                onPageChanged: (int index) {
                  setState(() {
                    _currentIndex = index;
                  });

                  if (widget.onIndexChanged != null) {
                    widget.onIndexChanged!(index);
                  }
                },
              )),
          Positioned(
            bottom: 10,
            child: SwipeIndicatorWidget(
                currentIndex: _currentIndex,
                itemCount: widget.itemCount,
                theme: widget.indicatorTheme),
          ),
        ],
      ),
    );
  }
}

class SwipeIndicatorTheme {
  final Color activeColor;
  final Color inactiveColor;
  final Color backgroundColor;

  SwipeIndicatorTheme({
    required this.activeColor,
    required this.inactiveColor,
    required this.backgroundColor,
  });
}
