import 'dart:async';
import 'dart:io';

import 'package:connectivity_plus/connectivity_plus.dart';

import '../../flutter_waff.dart';

/// Listen network status and broadcast [ConnectivityEvent] event via [eventBus].
///
/// ```
/// // Listen ConnectivityState
/// var connectivityEventStream = eventBus.on<ConnectivityEvent>().listen((event) {
///   if (event.state == ConnectivityState.WIFI_AVAILABLE) {
///     ...
///   }
/// });
///
/// // Start listener
/// ConnectivityListener.startup(['google.com', 'youtube.com']);
///
/// //
/// connectivityEventStream.cancel();
/// ```
class ConnectivityListener {
  static StreamSubscription? _listener;

  static ConnectivityState _lastState = ConnectivityState.NONE;

  /// Start network status listen (single)
  ///
  /// If [availabilityHosts] is List<String>, detect from the first one until
  /// it succeeds.
  ///
  /// If [availabilityHosts] is null, only [ConnectivityState.BLUETOOTH],
  /// [ConnectivityState.WIFI], [ConnectivityState.ETHERNET],
  /// [ConnectivityState.MOBILE], [ConnectivityState.NONE] three events will be
  /// broadcast.
  ///
  static void startup({List<String>? availabilityHosts}) async {
    if (_listener == null) {
      Log.d("ConnectivityListener: startup");

      // Send notification immediately when first execution and no network
      final res = enumFromString(
          ConnectivityState.values,
          enumToString(await (Connectivity().checkConnectivity()))
              .toUpperCase());
      if (res == ConnectivityState.NONE) {
        _lastState = res;
        Log.d("ConnectivityListener: Fire ConnectivityEvent $_lastState");
        eventBus.fire(ConnectivityEvent(res));
      }

      // Start listen
      _listener = Connectivity()
          .onConnectivityChanged
          .listen((ConnectivityResult result) async {
        checkState(availabilityHosts: availabilityHosts);
      });
    } else {
      Log.d("ConnectivityListener: Fire ConnectivityEvent $_lastState");
      eventBus.fire(ConnectivityEvent(_lastState));
    }
  }

  /// Sample usage to check current state.
  ///
  /// If [availabilityHosts] is List<String>, detect from the first one until
  /// it succeeds.
  ///
  /// If [availabilityHosts] is null, [ConnectivityListener.availabilityHosts]
  /// will be used. if both are null, only return [ConnectivityState.BLUETOOTH],
  /// [ConnectivityState.WIFI], [ConnectivityState.ETHERNET],
  /// [ConnectivityState.MOBILE], [ConnectivityState.NONE] three states.
  ///
  /// ```
  /// final state = await ConnectivityListener.checkState();
  /// // ------ or ------
  /// final state = await ConnectivityListener.checkState(
  ///   availabilityHosts:['google.com', 'youtube.com'],
  /// );
  /// // ------ or ------
  /// ConnectivityListener.startup(
  ///   availabilityHosts: ['google.com', 'youtube.com'],
  /// );
  /// final state = await ConnectivityListener.checkState();
  /// ```
  static Future<ConnectivityState> checkState(
      {List<String>? availabilityHosts}) async {
    final res = enumFromString(ConnectivityState.values,
        enumToString(await (Connectivity().checkConnectivity())).toUpperCase());

    _lastState = res;
    Log.d("ConnectivityListener: Fire ConnectivityEvent $_lastState");
    eventBus.fire(ConnectivityEvent(res));

    if (availabilityHosts != null && res != ConnectivityState.NONE) {
      var availabilityRes = await _Availability.check(res, availabilityHosts);
      if (availabilityRes != null) {
        _lastState = availabilityRes;
        Log.d("ConnectivityListener: Fire ConnectivityEvent $_lastState");
        eventBus.fire(ConnectivityEvent(availabilityRes));
      }
      return availabilityRes ?? res;
    }

    return res;
  }

  /// Stop listener
  static void stop() {
    _listener?.cancel();
    _listener = null;
  }
}

class _Availability {
  static String? _tokenKey;

  static Future<ConnectivityState?> check(
      ConnectivityState preState, List<String> hosts) async {
    var tokenKey = _createTokenKey();
    _tokenKey = tokenKey;

    for (int i = 0; i < hosts.length; i++) {
      try {
        Log.d(
            "ConnectivityListener: Lookup ${hosts[i]}, Token: $tokenKey, Action: Start");
        final result = await InternetAddress.lookup(hosts[i]);
        if (_tokenKey != tokenKey) {
          Log.d(
              "ConnectivityListener: Lookup ${hosts[i]}, Token: $tokenKey, Action: Skip token $_tokenKey");
          return null;
        }
        if (result.isNotEmpty && result[0].rawAddress.isNotEmpty) {
          Log.d(
              "ConnectivityListener: Lookup ${hosts[i]}, Token: $tokenKey, Action: AVAILABLE");
          return enumFromString(
              ConnectivityState.values, '${enumToString(preState)}_AVAILABLE');
        }

        Log.d(
            "ConnectivityListener: Lookup ${hosts[i]}, Token: $tokenKey, Action: NOT_AVAILABLE");
      } catch (error) {
        Log.e(
            "ConnectivityListener: Lookup ${hosts[i]}, Token: $tokenKey, Action: Error ${error.toString()}");
      }
    }

    return enumFromString(
        ConnectivityState.values, '${enumToString(preState)}_NOT_AVAILABLE');
  }

  static String _createTokenKey() {
    return DateTime.now().millisecondsSinceEpoch.toString();
  }
}

class ConnectivityEvent {
  final ConnectivityState state;

  ConnectivityEvent(this.state);
}

enum ConnectivityState {
  /// Device connected via bluetooth，but availability has not been detected.
  BLUETOOTH,

  /// Device connected via bluetooth，and the availability has been tested.
  BLUETOOTH_AVAILABLE,

  /// Device connected via bluetooth，and the availability test is not passed.
  BLUETOOTH_NOT_AVAILABLE,

  /// Device connected via Wi-Fi，but availability has not been detected.
  WIFI,

  /// Device connected via Wi-Fi，and the availability has been tested.
  WIFI_AVAILABLE,

  /// Device connected via Wi-Fi，and the availability test is not passed.
  WIFI_NOT_AVAILABLE,

  /// Device connected to ethernet network，but availability has not been detected.
  ETHERNET,

  /// Device connected to ethernet network，and the availability has been tested.
  ETHERNET_AVAILABLE,

  /// Device connected to ethernet network，and the availability test is not passed.
  ETHERNET_NOT_AVAILABLE,

  /// Device connected to cellular network，but availability has not been detected.
  MOBILE,

  /// Device connected to cellular network，and the availability has been tested.
  MOBILE_AVAILABLE,

  /// Device connected to cellular network，and the availability test is not passed.
  MOBILE_NOT_AVAILABLE,

  /// evice not connected to any network
  NONE,
}
