import 'dart:io';
import 'dart:ui';

import 'package:get/get.dart';

import '../services/db/storage_service.dart';
import '../waff_localizations.dart';

typedef TranslationKey = Map<String, String>;
typedef TranslationKeys = Map<String, TranslationKey>;

abstract class LocaleManage extends Translations {
  final List<LocaleManageSetting> settings;
  final StorageService storage;
  final String storageKey;

  // Current locale name
  late final Rx<String> currentName;

  LocaleManage({
    required this.settings,
    required this.storage,
    this.storageKey = 'localeName',
  }) : assert(settings.length > 0) {
    String? localeName = storage.getString(storageKey);
    if (!supportedLocales.keys.contains(localeName)) {
      localeName = findLocaleName(supportedLocales.keys, Platform.localeName);
    }

    currentName = Rx<String>(!supportedLocales.keys.contains(localeName)
        ? supportedLocales.keys.first
        : localeName!);
  }

  /// Current locale
  Locale get current {
    return supportedLocales[currentName.value]!;
  }

  /// Fallback locale
  Locale get fallbackLocale => supportedLocales[supportedLocales.keys.first]!;

  TranslationKeys? _keys;

  @override
  TranslationKeys get keys {
    if (_keys != null) {
      return _keys!;
    }

    _keys = {};
    for (int i = 0; i < settings.length; i++) {
      String rootName = settings[i].locale.languageCode;
      String fullName = settings[i].locale.languageCode;
      if (settings[i].locale.countryCode != null) {
        fullName =
            '${settings[i].locale.languageCode}_${settings[i].locale.countryCode}';
      }
      if (!_keys!.containsKey(rootName)) {
        _keys![rootName] = settings[i].translationKey;
      }
      if (!_keys!.containsKey(fullName)) {
        _keys![fullName] = settings[i].translationKey;
      }
    }

    return _keys!;
  }

  Map<String, Locale>? _supportedLocales;

  /// Get supported Locales
  Map<String, Locale> get supportedLocales {
    if (_supportedLocales != null) {
      return _supportedLocales!;
    }

    _supportedLocales = {};
    for (int i = 0; i < settings.length; i++) {
      String rootName = settings[i].locale.languageCode;
      String fullName = settings[i].locale.languageCode;
      if (settings[i].locale.countryCode != null) {
        fullName =
            '${settings[i].locale.languageCode}_${settings[i].locale.countryCode}';
      }
      if (!_supportedLocales!.containsKey(rootName)) {
        _supportedLocales![rootName] = settings[i].locale;
      }
      if (!_supportedLocales!.containsKey(fullName)) {
        _supportedLocales![fullName] = settings[i].locale;
      }
    }

    return _supportedLocales!;
  }

  /// Change locale
  void change(String? localeName) {
    if (supportedLocales.containsKey(localeName) &&
        localeName != currentName.value) {
      storage.setString(storageKey, localeName!);
      Get.updateLocale(supportedLocales[localeName]!);

      currentName.value = localeName;
    }
  }
}

class LocaleManageSetting {
  final Locale locale;
  final TranslationKey translationKey;

  LocaleManageSetting(this.locale, this.translationKey);
}
