import 'dart:async';
import 'dart:developer' as developer;

import 'package:logger/logger.dart';

import 'utils.dart';

typedef FutureOr<void> _OnCollectCallback(LogLevel level, List<String> lines);

enum LogLevel {
  verbose,
  debug,
  info,
  warning,
  error,
  wtf,
  nothing,
}

/// Logger tools, secondary packaging based on "logger/logger".
///
/// Through the [enableConsole] and [onCollect] parameters to control
/// the log output switch.
///
/// ```
/// Log.enableConsole = true;
/// Log.onCollect = (LogLevel level, List<String> lines) {
///   sendToNetwork(level, lines.join("\n"));
/// };
/// Log.i({"sample":"map"});
/// Log.d(true);
/// Log.e("Sample string");
/// ```
abstract class Log {
  static bool enableConsole = false;
  static _OnCollectCallback? onCollect;

  static Level _level = Level.error;
  static set logLevel(LogLevel level) {
    _level = enumFromString<Level>(Level.values, enumToString(level));
  }

  static Logger? _loggerInstance;
  static Logger get _logger {
    if (null == _loggerInstance) {
      var outputs = <LogOutput>[];
      if (enableConsole) {
        outputs.add(_ConsoleOutput());
      }

      if (onCollect != null) {
        outputs.add(_CollectorOutput(onCollect!));
      }

      _loggerInstance = Logger(
        level: _level,
        filter: ProductionFilter(),
        printer: _Printer(),
        output: MultiOutput(outputs),
      );
    }

    return _loggerInstance!;
  }

  /// Log a message at level [Level.verbose].
  static void v(dynamic message, [StackTrace? stackTrace]) {
    _logger.v(message, null, stackTrace);
  }

  /// Log a message at level [Level.debug].
  static void d(dynamic message, [StackTrace? stackTrace]) {
    _logger.d(message, null, stackTrace);
  }

  /// Log a message at level [Level.info].
  static void i(dynamic message, [StackTrace? stackTrace]) {
    _logger.i(message, null, stackTrace);
  }

  /// Log a message at level [Level.warning].
  static void w(dynamic message, [StackTrace? stackTrace]) {
    _logger.w(message, null, stackTrace);
  }

  /// Log a message at level [Level.error].
  static void e(dynamic message, [StackTrace? stackTrace]) {
    _logger.e(message, null, stackTrace);
  }

  /// Log a message at level [Level.wtf].
  static void wtf(dynamic message, [StackTrace? stackTrace]) {
    _logger.wtf(message, null, stackTrace);
  }
}

Map<Level, String> _levelNames = {
  Level.verbose: 'V',
  Level.debug: 'D',
  Level.info: 'I',
  Level.warning: 'W',
  Level.error: 'E',
  Level.wtf: 'WTF',
};

class _Printer extends PrettyPrinter {
  _Printer({
    stackTraceBeginIndex = 0,
    errorMethodCount = 30,
  }) : super(
          stackTraceBeginIndex: stackTraceBeginIndex,
          errorMethodCount: errorMethodCount,
        );

  @override
  List<String> log(LogEvent event) {
    var messageStr = stringifyMessage(event.message);

    String? stackTraceStr;
    if (event.stackTrace != null && errorMethodCount > 0) {
      stackTraceStr = formatStackTrace(event.stackTrace, errorMethodCount);
    }

    var errorStr = event.error?.toString();

    return _formatAndPrint(
      event.level,
      messageStr,
      errorStr,
      stackTraceStr,
    );
  }

  List<String> _formatAndPrint(
    Level level,
    String message,
    String? error,
    String? stacktrace,
  ) {
    List<String> buffer = [];

    DateTime date = DateTime.now();
    String time = '${date.hour}:${date.minute}:${date.second}';

    String prefix = '[${_levelNames[level]!}] [$time]';

    // format message
    for (var line in message.split('\n')) {
      buffer.add('$prefix $line');
    }

    if (error != null) {
      for (var line in error.split('\n')) {
        buffer.add('$prefix $line');
      }
    }

    if (stacktrace != null) {
      for (var line in stacktrace.split('\n')) {
        buffer.add('$prefix $line');
      }
    }

    return buffer;
  }
}

class _ConsoleOutput extends LogOutput {
  @override
  void output(OutputEvent event) {
    event.lines.forEach((element) {
      developer.log(element.replaceAll(RegExp(r'^\[\w+\]\s+'), ''),
          name: _levelNames[event.level]!);
    });
  }
}

class _CollectorOutput extends LogOutput {
  final _OnCollectCallback onCollect;

  _CollectorOutput(this.onCollect);

  @override
  void output(OutputEvent event) {
    this.onCollect(
      enumFromString<LogLevel>(LogLevel.values, enumToString(event.level)),
      event.lines,
    );
  }
}
