import 'package:flutter/material.dart';
import 'package:get/get.dart';

import '../services/db/storage_service.dart';

abstract class ThemeManage<C extends ThemeManageColors> {
  static const int LIGHT = 0;
  static const int DARK = 1;
  static const int SYSTEM = -1;

  static final Map<int, ThemeMode> themeModeChart = {
    LIGHT: ThemeMode.light,
    DARK: ThemeMode.dark,
    SYSTEM: ThemeMode.system,
  };

  final bool allowChangeThemeMode;
  final ThemeData light;
  final C lightColors;
  final ThemeData? dark;
  final C? darkColors;
  final StorageService storage;
  final String storageKey;

  late final Rx<ThemeMode> mode;

  ThemeManage({
    required this.allowChangeThemeMode,
    required this.light,
    required this.lightColors,
    this.dark,
    this.darkColors,
    required this.storage,
    this.storageKey = 'themeMode',
  }) : assert(!allowChangeThemeMode ||
            allowChangeThemeMode && dark != null && darkColors != null) {
    if (!allowChangeThemeMode) {
      mode = themeModeChart.values.first.obs;
    }
    int? themeModeId = storage.getInt(storageKey);
    mode = (themeModeChart.containsKey(themeModeId)
            ? themeModeChart[themeModeId]!
            : themeModeChart.values.first)
        .obs;
  }

  /// Current theme
  ThemeData getCurrentTheme() {
    if (!allowChangeThemeMode) {
      return light;
    }

    if (getCurrentBrightness() == Brightness.light) {
      return light;
    }

    return dark!;
  }

  /// Current colors
  C getCurrentColors() {
    if (!allowChangeThemeMode) {
      return lightColors;
    }

    if (getCurrentBrightness() == Brightness.light) {
      return lightColors;
    }

    return darkColors!;
  }

  /// Current brightness
  Brightness getCurrentBrightness() {
    if (mode.value == ThemeMode.light) {
      return Brightness.light;
    } else if (mode.value == ThemeMode.dark) {
      return Brightness.dark;
    }

    // Follow system
    if (Get.context != null) {
      return MediaQuery.of(Get.context!).platformBrightness;
    }

    return Brightness.light;
  }

  /// Change theme mode
  void changeMode(ThemeMode themeMode) {
    if (!allowChangeThemeMode) {
      return;
    }

    if (themeMode == ThemeMode.light) {
      storage.setInt(storageKey, LIGHT);
    } else if (themeMode == ThemeMode.dark) {
      storage.setInt(storageKey, DARK);
    } else if (themeMode == ThemeMode.system) {
      storage.setInt(storageKey, SYSTEM);
    }

    mode.value = themeMode;
  }
}

abstract class ThemeManageColors {}
